package be.jeffcheasey88.peeratcode.parser.java;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import be.jeffcheasey88.peeratcode.parser.java.CleanerPool.Cleaner;
import be.jeffcheasey88.peeratcode.parser.java.operations.OperationContainer;

public class Function extends OperationContainer{
	
	private static Pattern PATTERN = Pattern.compile("^(\\s*([^(]*)\\(([^)]*)\\)\\s*([^{]*)\\{)(.*)$");
	private static Pattern ANNOTATION_PATTERN = Pattern.compile("^(\\s*(@\\s*[^\\s]+\\s*\\^GENERIC_PARENTHESIS\\d+\\s+))");
	
	private List<Annotation> annotations;
	private int modifier;
	private String returnType;
	private String name;
	private List<Variable> parameters;
	private String exceptions;
	
	private boolean constructor;
	
	public Function(){
		this.annotations = new ArrayList<>();
		this.parameters = new ArrayList<>();
	}
	
	public int parse(String content, CleanerPool global, CleanerPool local) throws Exception{
		local = new CleanerPool(new Cleaner("GENERIC_PARENTHESIS", '(', ')'));
		content = local.clean(content);
		
		int annotationOffset = 0;
		
		Matcher matcher;
		while((matcher = ANNOTATION_PATTERN.matcher(content)).lookingAt()){
			String annot = matcher.group(1);
			annotationOffset+=local.unzip(annot, (s,p) -> s).length();
			content = content.substring(annot.length());
			Annotation annotation = new Annotation();
			annotation.parse(matcher.group(2), global, local);
			this.annotations.add(annotation);
		}
		
		content = local.unzipOne(local.clean(content), (s,p) -> s);
		matcher = PATTERN.matcher(content);
		matcher.matches();
		
		attribute(matcher.group(2));
		parameters(matcher.group(3)+";", global, local);
		this.exceptions = matcher.group(4).trim();
		
		local = local.group(new CleanerPool(new Cleaner("GENERIC_FUNCTION", '{', '}')));
		
		
		String zip = local.clean("{"+matcher.group(5));
		String body = local.unzipOne(zip, (s,p) -> s);
		String unzip = body.substring(1, body.indexOf('}'));

		super.parse(local.clean(unzip), global, local);
		
		return annotationOffset+local.unzip(matcher.group(1), (s,p) -> s).length()+local.unzip(unzip, ((s,p) -> s)).length()+1;
	}
	
	private static Pattern UNZIP_STICK = Pattern.compile("\\s+(?<e>[<|(|\\[|\"|'])");
	private static Pattern UNZIP_MAJ = Pattern.compile(">(?<e>[^>\\d,;(])");
	private static Pattern UNZIP_ARRAY = Pattern.compile("](?<e>[^\\[\\d,;])");
	
	private void attribute(String content) throws Exception{
		CleanerPool generic = new CleanerPool(
				new Cleaner("GENERIC_TYPE_",'<','>'),
				new Cleaner("GENERIC_ARRAY",'[',']'));
		String zip = generic.clean(content);
		String unzip = generic.unzip(zip, (value, pattern) -> {
			return value.replaceAll("\\s+", "");
		});
		
		unzip = UNZIP_STICK.matcher(unzip).replaceAll("${e}");
		unzip = UNZIP_MAJ.matcher(unzip).replaceAll("> ${e}");
		unzip = UNZIP_ARRAY.matcher(unzip).replaceAll("] ${e}");
		
		System.out.println("split space on "+unzip);
		
		Iterator<String> values = new ArrayIterator<>(unzip.split("\\s+"));
		String value = null;
		int modifier;
		while(values.hasNext() && ((value = values.next()).charAt(0) == '@')){
			Annotation annotation = new Annotation();
			annotation.parse(value, null, generic);
			this.annotations.add(annotation);
			System.out.println("parse annotation "+value);
		}
		System.out.println("out in("+values.hasNext()+" -> "+value.charAt(0)+") "+value);
		if((modifier = JavaParser.getModifier(value)) > 0){
			do{
				this.modifier+=modifier;
			}while(values.hasNext() && (modifier = JavaParser.getModifier(value = values.next())) > 0);
		}
		if(this.returnType == null){
			this.returnType = value;
			if(values.hasNext()) value = values.next();
			else constructor = true;
		}
		if(this.name == null){
			this.name = value;
		}
	}
	
	private void parameters(String content, CleanerPool global, CleanerPool local) throws Exception{
		if(content.length() == 1) return;
		boolean quote = false;
		do {
			Variable variable = new Variable();
			int index = variable.parse(content, global, local);
			this.parameters.add(variable);
			content = content.substring(index);
			quote = content.startsWith(",");
			if(quote) content = content.substring(1);
		}while(quote);
	}
	
	@Override
	public void build(ArrayBuffer<String> buffer, int tab) throws Exception{
		for(Annotation annotation : this.annotations){
			annotation.build(buffer, tab);
			buffer.add("");
		}
		
		boolean empty = getChilds().size() == 0;
		
		super.build(buffer, tab);
		buffer.append((s) -> s+=Modifier.toString(modifier)+" "+(constructor ? "" : returnType+" ")+name+"(");
		boolean first = true;
		for(Variable variable : this.parameters){
			for(Annotation annotation : variable.getAnnotations()){
				annotation.build(buffer, 0);
				buffer.append((s) -> s+=" ");
			}
			if(first) first = false;
			else{
				buffer.append((s) -> s+=",");
			}
			buffer.append((s) -> s+=variable.getType()+" "+variable.getName());
		}
		buffer.append((s) -> s+=") "+exceptions+"{"+((empty ? "}":"")));
		buffer.add("");
		
		if(empty) return;
		
		for(JavaElement child : getChilds()) child.build(buffer, tab+1);
		
		super.build(buffer, tab);
		buffer.append((s) -> s+="}");
		buffer.add("");
	}
	
}
